	.title	pdp8dis - pdp8 mnemonic disassembler
	.module	pdp8dis
	.radix	d

  ; CORESYS09 references

OUTB2O	=	0xC021	; output byte as 2 hex characters
outch	=	0xC024	; print character in [A] - not destroyed
OUTS	=	0xC03C	; output space
badsyn	=	0xC0AF	; bad syntax
gtrang	=	0xC7F6	; get range
nomore	=	0xC0C8	; return to CORESYS-09

ranglo	=	0x21B5	; low address from "gtrang"
ranghi	=	0x21B7	; high address


	.area	WORKPG	(ABS,OVR)

  ; storage for disassembly (accessed via [X] pointer)

opc8	= 	0	; PDP8 opcode
coradr	=	2	; current PDP8 disassembly core adress
endadr	=	4	; last PDP8 disassembly core address
memadr	=	6	; current 6809 disassembly real memory address

disas:	.rmb	8	; allocate storage accessed through [X]

CORE	=	0x2300	; start address of PDP8 core in 6809 RAM space
CR	=	0x0D
LF	=	0x0A


	.area	PDP8DIS (ABS,OVR)

;     *-------------------------------*
;     #   entry point for execution   #
;     *-------------------------------*
;
; commands:
;   pdp8d adr1:adr2    - disassemble from address1 to address2
;   pdp8d adr1!number  - disassemble from address1 <number> words
;   pdp8d adr1         - disassemble address1
;
;   NOTE: the addresses are PDP8 addresses and
;	  are specified in the active CORESYS "INBASE" setting (hex/octal)

pdp8d:	jsr	gtrang		; get begin and end address
	lble	badsyn
	ldx	#disas		; load pointer to local storage
	ldd	ranghi		; set end address
	std	endadr,x
	ldd	ranglo		; get start address
loop:	bsr	dis8as		; do one line disassembly
	cmpd	endadr,x	; disassembly done ?
	blo	loop		; n - continue next instruction
	jmp	nomore



;    *--------------------------------------*
;    *   disassemble one PDP8 instruction   *
;    *--------------------------------------*
;
;  call   : [D] = PDP8 address of the instruction opcode
;	    [X] = pointer to local storage
;  return : [D] : points to next PDP8 instruction address
;	    [X] : not changed
;
;	    complete disassembly line printed to active console
;		- hexadecimal values have a leading "$"
;		- octal values have a leading "o"
;	    output format:
;		- hex real 6809 address
;		- in parenthesis hex and octal PDP8 address followed by a colon
;		- instruction in hex and octal
;		- instruction mnemonic, 3 characters, space and "i" or space
;		- operand address or IOT/OPR information

dis8as:	std	coradr,x	; save working PDP8 address
	aslb			; calculate real (6809) address
	rola
	addd	#CORE
	std	memadr,x
	tfr	d,y
	ldd	,y		; get PDP8 opcode
	std	opc8,x
	ldy	#stline
	jsr	prtdis		; start new line with hex indicator
	ldd	memadr,x
	jsr	prt4hx		; print real memory address
	jsr	OUTS
	jsr	openbr		; "("
	ldd	coradr,x
	jsr	prt4hx		; print core address hex
	jsr	hexoct		; print hex-octal separator "/ o"
	ldd	coradr,x
	jsr	oct4		; print core address octal
	jsr	closbr		; ")"
	ldy	#adrsep
	jsr	prtdis		; print address - opcode separator
	ldd	opc8,x
	jsr	prt4hx		; print instruction hex format
	jsr	hexoct		; print hex-octal separator "/ o"
	ldd	opc8,x
	jsr	oct4		; and in octal
	jsr	pr2sp		; print 2 spaces

;  process i / z / operand fields

	lda	opc8,x		; get msb of instruction
	tfr	a,b
	aslb
	andb	#0b00011100
	ldy	#Opstbl
	leay	b,y		; point to mnemonic text
	anda	#0b00001110	; check for OPR : modified output
	cmpa	#0b00001110
	beq	opropc		; - OPR: print correct mnemonic for groups 1 & 2
	jsr	prtdis		; print AND, TAD, ISZ, DCA, JMS, JMP, IOT mnemonic
	jsr	OUTS		; followed by 1 space instruction separator
	cmpa	#0b00001100	; check for IOT : different output format
	beq	iotopc		; - IOT: print device # and operation code


;    11  10  9   8   7   6   5   4   3   2   1   0
;    ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ 
;   |   |   |   |   |   |   |   |   |   |   |   |   |  i - [0=direct, 1=indirect]
;   |_ _|__ |__ |__ |__ |___|___|__ |__ |__ |__ |__ |  z - [0=page 0, 1=current page
;   :           :   :   :                           :
;   :  opcode   : i : z :           address         :  address - 12-bit word in page
;
; output:
;   $xxxx ($0xxx / oxxxx):  $0xxx / oxxxx  TAD i  $xxxx / oxxxx  ($xxxx)

	lda	#' 		; assume not indirect addressing
	ldb	opc8,x
	bitb	#0b00000001	; check "i" bit
	beq	10$
	lda	#'i
10$:	jsr	outch		; print space or "i"
	bsr	pr2sp		; 2 spaces
	ldb	opc8+1,x	; check page zero bit
	bmi	20$
	clr	opc8,x		; set address in page zero
	bra	30$
20$:	clra
	andb	#0b01111111	; offset in current page
	std	opc8,x
	ldd	coradr,x
	andb	#0b10000000	; get page
	addd	opc8,x
	std	opc8,x		; set address in current page
30$:	ldd	opc8,x
	bsr	prt4hx		; print operand address hex
	bsr	hexoct		; print hex-octal separator "/ o"
	ldd	opc8,x
	jsr	oct4		; print core address octal
	bsr	pr2sp		; print 2 spaces
	bsr	openbr		; "("
	ldd	opc8,x
	aslb
	rola
	addd	#CORE
	bsr	prt4hx		; print operand real address hex
	bsr	closbr		; ")"
	bra	done


;  print IOT
;         _ _ _ _|_ _ _ _ _ _ _ _
;        |1|1|0|_|_|_|_|_|_|_|_|_|
;        |     |           |     |
;        |     |   device  | op  |
;
; output:
;   $4006 ($0E83 / o7203):  $0C33 / o6063  IOT    dvc = $06  op = 3

iotopc:	ldy	#device
	bsr	prtdis
	ldd	opc8,x		; get device #
	asra
	rorb
	asrb
	asrb
	andb	#0b00111111
	tfr	b,a
	jsr	oct2		; print device # as 2 octal digits
	ldy	#op
	bsr	prtdis
	lda	opc8+1,x
	jsr	oct1		; print operation
	bra	done


;  print OPR
;
;    Group 1 |1|1|1|0|_|_|_|_|_|_|_|_|
;    Group 2 |1|1|1|1|_|_|_|_|_|_|_|0|
;    Group 3 |1|1|1|1|_|_|_|_|_|_|_|1|
;
; output:
;   $xxxx ($0xxx / oxxxx):  $0xxx / oxxxx  CLA CLL
;   $xxxx ($0xxx / oxxxx):  $0xxx / oxxxx  HLT
;   $xxxx ($0xxx / oxxxx):  $0xxx / oxxxx  OPR    grp 3, bits: oxxx

opropc:	ldb	opc8,x
	asrb			; group 1 ?
	bcc	grp1		; y -
	ldb	opc8+1,x
	asrb			; group 2 ?
	lbcc	grp2		; y -
	ldy	#group3
	bsr	prtdis		; print "OPR    grp 3, bits: o"
	ldd	opc8,x
	jsr	oct3		; print bits as 3 octal digits

done:	ldd	coradr,x
	addd	#1
	std	coradr,x
	rts



;  print routines for the output line

closbr:	lda	#')
	bra	prchr
openbr:	jsr	OUTS
	lda	#'(
prchr:	jmp	outch

pr2sp:	jsr	OUTS		; print 2 spaces
	jmp	OUTS

hexoct:	ldy	#hx_oct		; print " / o"   and fall through



;  prtdis - print text string
;
;    call   : [Y] = pointer to start of string, ended *NULL*
;    return : [A] - preserved
;	      [Y] - points to the location following the NULL

prtdis:	pshs	a
	lda	,y+		; get char
	beq	prtx		; done -
prtd:	bsr	prchr		; print char ASCII
	lda	,y+
	bne	prtd
prtx:	puls	a,pc



;  prt4hx - print [D] as 4 hex characters with a leading "$"

prt4hx:	pshs	a
	lda	#'$
	bsr	prchr		; print leading "$"
	puls	a
	jsr	OUTB2O		; print [A] as 2 hex char
	tfr	b,a
	jmp	OUTB2O		; print [B] as 2 hex and RTS




; OPR group 1 disassembly
;
;    [B] =  |_|_|_|_|_|_|_|_|
;            1			-  CLA - clear AC
;              1		-  CLL - clear the L bit
;                1		-  CMA - ones complement AC
;                  1		-  CML - complement L bit
;                          1	-  IAC - increment <AC>
;                    1 0 0	-  RAR - rotate <L,AC> right
;                    0 1 0	-  RAL - rotate <L,AC> left
;                    1 0 1	-  RTR - rotate <L,AC> right twice
;                    0 1 1	-  RTL - rotate <L,AC> left twice
;                    0 0 1	-  6-bit "byte" swap AC (on pdp8/e)

grp1:	ldu	#0		; count # mnemonics printed
	ldb	opc8+1,x	; get lower 8 bits of group 1 OPR instruction
	aslb
	bcc	10$
	ldy	#CLA
	bsr	prtgp1		; print mnemonic with optional [U]-flag leading space
10$:	aslb
	bcc	20$
	ldy	#CLL
	bsr	prtgp1
20$:	aslb
	bcc	30$
	ldy	#CMA
	bsr	prtgp1
30$:	aslb
	bcc	40$
	ldy	#CML
	bsr	prtgp1
40$:	ldb	opc8+1,x
	asrb
	bcc	50$
	ldy	#IAC
	bsr	prtgp1
50$:	ldb	opc8+1,x
	aslb
	andb	#0b00011100

    ;  0 = 0 0 0 0 0  - *?*
    ;  1 = 0 0 1 0 0  - swap (on pdp8/e)
    ;  2 = 0 1 0 0 0  - RAL
    ;  3 = 0 1 1 0 0  - RTL
    ;  4 = 1 0 0 0 0  - RAR
    ;  5 = 1 0 1 0 0  - RTR
    ;  6 = 1 1 0 0 0  - RAR RAL
    ;  7 = 1 1 1 0 0  - RTR RTL

	ldy	#GP1TB
	leay	b,y
	cmpb	#0b00011100	; dirty: last 2 mnemonics more than 4 chars
	bne	grp1p		; it's the first one: pointer ok
	leay	4,y		; next (last) one: correct the pointer
grp1p:	bsr	prtgp1		; print mnemonic
	jmp	done




    ; prtgp1 - print text (group 1 mnomonic) with optional leading space
    ;
    ;   call  : [Y] = pointer to text string
    ;		[U] = leading space flag: 0 = no space  //  !0 = space
    ;   return: [U] incremented by +1

prtgp1:	lda	#' 		; set leading space



    ; prtgp2 - print text (group 2 mnomonic) with optional leading character
    ;
    ;   call  : [Y] = pointer to text string
    ;		[A] = leading character to print
    ;		[U] = leading character flag: 0 = no char  //  !0 = char
    ;   return: [U] incremented by +1
    ;		[A] not changed

prtgp2:	leau	1,u		; one mnemonic will be printed: next one needs leading space
	cmpu	#1
	beq	prts2
	jsr	outch		; leading character needed
prts2:	bra	prtdis		; print string and RTS




; OPR group 2 disassembly
;
;    [B] =  |_|_|_|_|_|_|_|0|
;              1     0		-  SMA - skip on AC < 0      \
;                1   0		-  SZA - skip on AC = 0       > OR group
;                  1 0		-  SNL - skip on L != 0      /
;              0 0 0 1		-  SKP - skip unconditional
;              1     1		-  SPA - skip on AC >= 0     \
;                1   1		-  SNA - skip on AC != 0      > AND group
;                  1 1		-  SZL - skip on L = 0       /
;            1			-  CLA - clear AC
;                      1	-  OSR - or switches with AC
;                        1	-  HLT - halt

grp2:	ldu	#0		; count # mnemonics printed
	ldb	opc8+1,x	; get lower 8 bits of group 2 OPR instruction
	bpl	10$
	ldy	#CLA
	bsr	prtgp2		; 1st mnemonic, no leading character
10$:	bitb	#0b00000100
	beq	20$
	ldy	#OSR
	lda	#' 		; optional leading space
	bsr	prtgp2
20$:	andb	#0b01111000
	cmpb	#0b00001000
	bne	30$
	ldy	#SKP
	lda	#' 		; optional leading space
	bra	grp2dn
30$:	ldy	#GP2OR		; assume OR group: set pointer to group 2 OR codes
	lda	#'+		;		   set OR "+" separator
	bitb	#0b00001000	; correct ?
	beq	grp20		; y -
	ldy	#GP2AND		; n - set pointer to group 2 AND instruction group
	lda	#'.		;     set AND "." separator
grp20:	aslb
grp21:	aslb
	bcs	grp2_1
	leay	4,y		; skip first: bump text pointer	
	bra	grp22
grp2_1:	pshs	a		; save AND/OR separator: first is always a space
	bsr	prtgp1		; print first instruction (opt. always leading space)
	puls	a		; restore AND/OR separator
grp22:	aslb
	bcs	grp2_2
	leay	4,y		; skip: bump text pointer	
	bra	grp23
grp2_2:	bsr	prtgp2		; print instruction (and optional separator token)
grp23:	aslb
	bcc	grp2h
grp2dn:	bsr	prtgp2		; print instruction (and optional separator token)
grp2h:	ldb	opc8+1,x
	bitb	#0b00000010
	beq	grp2x
	ldy	#HLT
	lda	#' 		; optional leading space
	bsr	prtgp2
grp2x:	jmp	done




    ; oct4 - print 12-bit value [D] in 4 digits octal format
    ; oct3 - print  9-bit value [D] in 3 digits octal format
    ; oct2 - print  6-bit value [A] in 2 digits octal format
    ; oct1 - print  3-bit value [A] in 1 digits octal format

    ; [A,B] =  0|0|0|0|_|_|_|_|   |_|_|_|_|_|_|_|_|

oct4:	pshs	a
	asra			; first 3 bits
	bsr	oct1		; print 1st digit
	puls	a
oct3:	pshs	b
	aslb			; next 3 bits
	rola
	aslb
	rola
	bsr	oct1		; print 2nd digit
	puls	a
oct2:	pshs	a
	asra
	asra
	asra	
	bsr	oct1		; print 3rd digit
	puls	a		; print 4th digit

oct1:	anda	#7
	adda	#'0
	jmp	prchr		; print digit and RTS



   ; PDP8 opcodes lookup table

Opstbl:	.asciz	"AND"
	.asciz	"TAD"
	.asciz	"ISZ"
	.asciz	"DCA"
	.asciz	"JMS"
	.asciz	"JMP"
	.asciz	"IOT"
; --	.asciz	"OPR"		; not referenced / separate case


   ; OPR mnemonics - group 1

CLL:	.asciz	"CLL"
CMA:	.asciz	"CMA"
CML:	.asciz	"CML"
IAC:	.asciz	"IAC"

    ;  0 = 0 0 0 0 0  - *?*
    ;  1 = 0 0 1 0 0  - swap (on pdp8/e)
    ;  2 = 0 1 0 0 0  - RAL
    ;  3 = 0 1 1 0 0  - RTL
    ;  4 = 1 0 0 0 0  - RAR
    ;  5 = 1 0 1 0 0  - RTR
    ;  6 = 1 1 0 0 0  - RAR RAL
    ;  7 = 1 1 1 0 0  - RTR RTL

GP1TB:	.asciz	"*?*"		; 0 0 0
	.asciz	"BSW"		; 0 0 1
	.asciz	"RAL"		; 0 1 0
	.asciz	"RTL"		; 0 1 1
	.asciz	"RAR"		; 1 0 0
	.asciz	"RTR"		; 1 1 0
	.asciz	"RAR RAL"	; 1 1 0  dirty: different size
	.asciz	"RTR RTL"	; 1 1 1    handled in the code


   ; OPR mnemonics - group 2

CLA:	.asciz	"CLA"
HLT:	.asciz	"HLT"
OSR:	.asciz	"OSR"
SKP:	.asciz	"SKP"
GP2OR:
SMA:	.asciz	"SMA"		; OR group
SZA:	.asciz	"SZA"
SNL:	.asciz	"SNL"
GP2AND:
SPA:	.asciz	"SPA"		; AND group
SNA:	.asciz	"SNA"
SZL:	.asciz	"SZL"


stline:	.fcb	CR,LF
	.asciz	"  "
hx_oct:	.asciz	" / o"
adrsep:	.asciz	":   "
device:	.asciz	"   dvc = o"
op:	.asciz	" op = "
group3:	.asciz	"OPR    grp 3, bits: o"

	.end